#define _GNU_SOURCE
#include <stdint.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <strings.h>
#include <smb2/smb2.h>
#include <smb2/libsmb2.h>
#include <smb2/libsmb2-raw.h>

#ifndef ARRAY_SIZE
#define ARRAY_SIZE(x) (sizeof(x) / sizeof(x[0]))
#endif

// This function is declared in libsmb2-private.h but we want to be able to use it
void smb2_set_error(struct smb2_context *smb2, const char *error_string, ...);

static size_t strlcpy(char *dst, const char *src, size_t maxlen)
{
	const size_t srclen = strlen(src);
	if (srclen + 1 < maxlen) {
		memcpy(dst, src, srclen + 1);
	} else if (maxlen != 0) {
		memcpy(dst, src, maxlen - 1);
		dst[maxlen - 1] = '\0';
	}
	return srclen;
}

// Prints out a SMB2 error message similar to perror(3)
static void smb2_perror(struct smb2_context *smb2, const char *msg)
{
	fprintf(stderr, "%s: %s\n", msg, smb2_get_error(smb2));
}

// Returns the dirname portion of a SMB path, similar to dirname(3).
// WARNING: This function is not thread-safe as it uses a static buffer.
static const char *smb2_dirname(const char *path)
{
	static char buf[1024];

	const char *pch = strrchr(path, '/');
	if (pch == NULL)
		return "";

	strlcpy(buf, path, pch - path + 1);
	return buf;
}

// Returns the basename portion of a SMB path, similar to basename(3).
static const char *smb2_basename(const char *path)
{
	const char *pch = strrchr(path, '/');
	if (pch == NULL)
		return path;
	else
		return pch + 1;
}

// Retrieves the SMB file id for a path.  Returns 0 on success, non-zero on failure.
static int smb2_get_fileid(struct smb2_context *smb2, const char *path, uint64_t *fileid)
{
	// It would be more efficient to get the fileid by sending a SMB_COM_NT_CREATE_ANDX
	// (0xA2) request and parsing its reply, which should map to smb2_create_request
	// and smb2_create_reply, and from there to smb2_open() and smb2_get_file_id().
	// However, I can't figure out why smb2_get_file_id() isn't returning a valid
	// fileid.
	//
	// Instead, I'm using SMB2_FILE_FULL_DIRECTORY_INFORMATION by way of smb2_opendir()
	// and smb2_readdir() which returns a file_id that I can parse correctly.  The
	// file id was confirmed by ensuring it matches `fsutil file queryfileid \\a\b\c`.
	struct smb2dir *dir;
	int found = 0;

	*fileid = 0;
	if ((dir = smb2_opendir(smb2, smb2_dirname(path))) != NULL) {
		struct smb2dirent *dirent;

		while ((dirent = smb2_readdir(smb2, dir)) != NULL && !found) {
			if (strcmp(dirent->name, smb2_basename(path)) == 0) {
				*fileid = dirent->st.smb2_ino;
				found = 1;
			}
		}

		smb2_closedir(smb2, dir);
	}

	if (!found) {
		smb2_set_error(smb2, "File not found");
	}
	return (found ? 0 : -1);
}

// Parses a URL string into a smb2_url.  Enhances smb2_parse_url to support
// URLs of the form //hostname/share/path in addition to smb://...
static struct smb2_url *smb2_parse_url_ex(struct smb2_context *smb2, const char *urlstr)
{
	struct smb2_url *url;
	char *buf;
	int ret;

	if ((url = smb2_parse_url(smb2, urlstr)) != NULL) {
		return url;
	}

	if ((ret = asprintf(&buf, "smb:%s", urlstr)) == -1) {
		smb2_set_error(smb2, "Out of memory");
		return NULL;
	}

	url = smb2_parse_url(smb2, buf);
	free(buf);
	return url;
}

static int queryfileid_subcmd(const char* argv0, int argc, char* argv[])
{
	struct smb2_context *smb2;
	struct smb2_url *url;
	uint64_t fileid;
	int ret = 1;

	if (argc != 1) {
		fprintf(stderr, "Usage: %s file queryFileID <filename>\n", basename(argv0));
		fprintf(stderr, "   Eg: %s file queryFileID //127.0.0.1/share/file.txt\n", basename(argv0));
		goto out;
	}

	smb2 = smb2_init_context();
	if (smb2 == NULL) {
		fprintf(stderr, "Failed to init context\n");
		goto out;
	}

	if ((url = smb2_parse_url_ex(smb2, argv[0])) == NULL) {
		smb2_perror(smb2, "smb2_parse_url");
		goto out2;
	}

	if (smb2_connect_share(smb2, url->server, url->share, NULL) != 0) {
		smb2_perror(smb2, "smb2_connect_share");
		goto out3;
	}

	if (smb2_get_fileid(smb2, url->path, &fileid)) {
		smb2_perror(smb2, "smb2_get_fileid");
		goto out4;
	}

	printf("0x%016lx\n", fileid);
	ret = 0;

out4:
	smb2_disconnect_share(smb2);
out3:
	smb2_destroy_url(url);
out2:
	smb2_destroy_context(smb2);
out:
	return ret;
}

static void print_usage(FILE* fp)
{
	fprintf(fp, "--- FILE Commands Supported ----\n");
	fprintf(fp, "\n");
	fprintf(fp, "queryFileID              Queries the file ID of the specified file\n");
}

int file_subcmd(const char* argv0, int argc, char* argv[])
{
	if (argc < 1) {
		print_usage(stderr);
		return 1;
	}

	if (strcasecmp(argv[0], "queryfileid") == 0) {
		return queryfileid_subcmd(argv0, argc - 1, argv + 1);
	} else {
		fprintf(stderr, "%s is an invalid parameter\n\n", argv[0]);
		print_usage(stderr);
		return 1;
	}
}
